-- name: Sharen's Animation Overhaul v0.1
-- description: A total re-do of a mayority of Mario's animations, complete with extra flairs such as sounds, open hands, etc! \n\nAuthor: Sharen

-- optimization
local hook_mod_menu_checkbox, math_min, math_max, approach_f32, math_floor, atan2s, set_anim_to_frame, smlua_anim_util_set_animation, is_anim_past_frame, play_sound, hook_event =
hook_mod_menu_checkbox, math.min, math.max, approach_f32, math.floor, atan2s, set_anim_to_frame, smlua_anim_util_set_animation, is_anim_past_frame, play_sound, hook_event

gMarioDetails = {}

for i = 0, MAX_PLAYERS - 1 do
    gMarioDetails[i] = {}
    d = gMarioDetails[i]
    s = gPlayerSyncTable[i]

    d.torsoAngle = {x = 0, y = 0, z = 0}

    s.newAnims = true
    s.useUnusedEyes = true
    s.airStrafeMove = true
    s.handStates = true
    s.divePitch = true
end

local mod_enable_cmd = function(_, value)
    gPlayerSyncTable[0].newAnims = value
end

local unused_eyes_cmd = function(_, value)
    gPlayerSyncTable[0].useUnusedEyes = value
end

local air_strafe_cmd = function(_, value)
    gPlayerSyncTable[0].airStrafeMove = value
end

local hand_states_cmd = function(_, value)
    gPlayerSyncTable[0].handStates = value
end

local dive_pitch_cmd = function(_, value)
    gPlayerSyncTable[0].divePitch = value
end

hook_mod_menu_checkbox("Use new animations", gPlayerSyncTable[0].newAnims, mod_enable_cmd)
hook_mod_menu_checkbox("Use unused eyes states", gPlayerSyncTable[0].useUnusedEyes, unused_eyes_cmd)
hook_mod_menu_checkbox("Move body when strafing", gPlayerSyncTable[0].airStrafeMove, air_strafe_cmd)
hook_mod_menu_checkbox("Open or grow hands in certain animations", gPlayerSyncTable[0].handStates, hand_states_cmd)
hook_mod_menu_checkbox("Dynamic pitch while diving", gPlayerSyncTable[0].divePitch, dive_pitch_cmd)

-- i dont use the MARIO_ANIM constants here so that i can know which number belongs to which animation in the animations file
local registeredAnims = {
    [194] = true, -- first person
    [195] = true, -- idle 1
    [196] = true, -- idle 2
    [197] = true, -- idle 3
    [114] = true, -- run
    [72] = true, -- walk
    [15] = true, -- skid
    [16] = true, -- skid end
    [188] = true, -- turn around
    [189] = true, -- turn around end
    [77] = true, -- jump
    [78] = true, -- jump land
    [80] = true, -- double jump rise
    [76] = true, -- double jump fall
    [75] = true, -- double jump land
    [193] = true, -- triple jump
    [192] = true, -- triple jump land
    [4] = true, -- backflip
    [191] = true, -- sideflip
    [190] = true, -- sideflip land
    [86] = true, -- freefall
    [87] = true, -- freefall land
    [152] = true, -- crouch
    [150] = true, -- crouch end
    [151] = true, -- crouch start
    [19] = true, -- long jump
    [17] = true, -- long jump land
    [20] = true, -- slow long jump
    [18] = true, -- slow long jump land
    [153] = true, -- crouch waddle (orignally crawling)
    [155] = true, -- crouch waddle start
    [154] = true, -- crouch waddle "end"
    [203] = true, -- wall jump
    [204] = true, -- wall slide
}

local deg_to_hex = function(x)
    return x * 0x10000 / 360
end

local hex_to_deg = function(x)
    return x * 360 / 0x10000
end

local clamp = function(x, min, max)
    return math_min(max, math_max(min, x))
end

local in_between = function(x, min, max, included)
    if included then
        if x >= min and x <= max then
            return true
        end
    end 

    if x > min and x < max then
        return true
    end

    return false
end

function s16(x)
    x = (math_floor(x) & 0xFFFF)
    if x >= 32768 then return x - 65536 end
    return x
end

---@param m MarioState
local turn_mario_head = function(m, angle, min, max)
    local s = gPlayerSyncTable[m.playerIndex]
    local mBody = m.marioBodyState

    mBody.headAngle.y = approach_f32(mBody.headAngle.y, clamp(angle, deg_to_hex(min), deg_to_hex(max)), 
    deg_to_hex(12), deg_to_hex(12))

    if not in_between(angle, deg_to_hex(min - 35), deg_to_hex(max + 35), true) and m.character.type ~= CT_TOAD and s.useUnusedEyes then
        if angle < 0 then
            mBody.eyeState = MARIO_EYES_LOOK_LEFT
        else
            mBody.eyeState = MARIO_EYES_LOOK_RIGHT
        end
    end
end

---@param m MarioState
local update_mario_body = function(m)
    local s = gPlayerSyncTable[m.playerIndex]
    local d = gMarioDetails[m.playerIndex]
    local id = m.marioObj.header.gfx.animInfo.animID
    local frame = m.marioObj.header.gfx.animInfo.animFrame
    local mBody = m.marioBodyState

    -- hands

    local openHandCond = {
        [15] = true,
        [16] = frame < 14,
        [80] = frame >= 2,
        [76] = true,
        [193] = not in_between(frame, 14, 21, false),
        [192] = in_between(frame, 4, 27, false),
        [4] = not in_between(frame, 17, 25, false),
        [191] = not in_between(frame, 9, 16, false),
        [86] = true,
        [19] = true,
        [17] = frame < 5,
        [20] = true,
        [18] = frame < 4,
        [203] = m.prevAction == ACT_AIR_HIT_WALL and true or frame <= 4,
        [136] = true,
        [137] = true,
    }

    if s.handStates then
        if openHandCond[id] then
            mBody.handState = MARIO_HAND_OPEN
        elseif id == 204 then
            mBody.handState = MARIO_HAND_RIGHT_OPEN
        end

        if id == 77 and in_between(frame, 3, 13, true) then
            mBody.punchState = (0 << 6) | 3
        end
    end

    -- eyes during idle

    if m.character.type ~= CT_TOAD and s.useUnusedEyes then
        if (id == 196 and in_between(frame, 4, 22, true)) or (id == 152 and in_between(frame, 2, 30, true)) then
            mBody.eyeState = MARIO_EYES_LOOK_LEFT
        elseif id == 188 or (id == 189 and frame <= 4) or (id == 195 and in_between(frame, 4, 22, true)) or
        (id == 152 and in_between(frame, 31, 60, true)) then
            mBody.eyeState = MARIO_EYES_LOOK_RIGHT
        end
    end

    -- dive pitch

    if m.action == ACT_DIVE and s.divePitch then
        m.faceAngle.x = atan2s(m.forwardVel, m.vel.y)
    end

    -- body tilting

    local lookOnStrafeActs = {
        [ACT_JUMP] = true,
        [ACT_DOUBLE_JUMP] = true,
        [ACT_TRIPLE_JUMP] = frame >= 28,
        [ACT_BACKFLIP] = frame >= 28,
        [ACT_WALL_KICK_AIR] = true,
        [ACT_LONG_JUMP] = true,
        [ACT_SIDE_FLIP] = frame >= 18,
        [ACT_FREEFALL] = true,
        [ACT_FORWARD_ROLLOUT] = m.actionState == 2,
        [ACT_BACKWARD_ROLLOUT] = m.actionState == 2,
    }

    mBody.allowPartRotation = 1
    local intendedDYaw = s16(m.intendedYaw - m.faceAngle.y)

    if m.action == ACT_WALKING and id == 114 and s.newAnims then
        if m.forwardVel < 28 then
            d.torsoAngle.x = approach_f32(d.torsoAngle.x, deg_to_hex(32), deg_to_hex(4.5), deg_to_hex(4.5))
        else
            d.torsoAngle.x = approach_f32(d.torsoAngle.x, deg_to_hex(12), deg_to_hex(1), deg_to_hex(1))
        end

        mBody.torsoAngle.x = d.torsoAngle.x
        mBody.headAngle.x = -mBody.torsoAngle.x * 0.65
        m.marioObj.header.gfx.angle.x = m.marioObj.oMarioWalkingPitch + mBody.torsoAngle.x * 0.25

    elseif lookOnStrafeActs[m.action] and s.airStrafeMove then
        local minAngle = (id == 77 and frame <= 13) and -20 or -60
        local lookAngle = atan2s(m.vel.z, m.vel.x) ~= 0 and s16(atan2s(m.vel.z, m.vel.x) - m.faceAngle.y) or 0

        turn_mario_head(m, lookAngle, minAngle, 60)

        local torsoXTarget = math_min(s16(6144 * (m.intendedMag / 32) * coss(intendedDYaw)) * 0.4, 0)
        local torsoZTarget = s16(-4096 * (m.intendedMag / 32) * sins(intendedDYaw)) * 0.4

        mBody.torsoAngle.x = approach_f32(mBody.torsoAngle.x, torsoXTarget, deg_to_hex(2), deg_to_hex(2))
        mBody.torsoAngle.y = 0
        mBody.torsoAngle.z = approach_f32(mBody.torsoAngle.z, torsoZTarget, deg_to_hex(2), deg_to_hex(2))

        mBody.headAngle.x = -mBody.torsoAngle.x
        mBody.headAngle.z = -mBody.torsoAngle.z
    else
        d.torsoAngle.x = 0
        d.torsoAngle.y = 0
        d.torsoAngle.z = 0
        mBody.allowPartRotation = 0
    end
end

---@param m MarioState
local active_player = function(m)
    local np = gNetworkPlayers[m.playerIndex]
    if m.playerIndex == 0 then
        return true
    end
    if not np.connected then
        return false
    end
    if np.currCourseNum ~= gNetworkPlayers[0].currCourseNum then
        return false
    end
    if np.currActNum ~= gNetworkPlayers[0].currActNum then
        return false
    end
    if np.currLevelNum ~= gNetworkPlayers[0].currLevelNum then
        return false
    end
    if np.currAreaIndex ~= gNetworkPlayers[0].currAreaIndex then
        return false
    end
    return true
end

local prevId = -1

---@param m MarioState
local mario_update = function(m)
    if not active_player(m) then
        return
    end

    local s = gPlayerSyncTable[m.playerIndex]
    local id = m.marioObj.header.gfx.animInfo.animID
    local frame = m.marioObj.header.gfx.animInfo.animFrame

    if registeredAnims[id] and s.newAnims then
        if id == 77 and m.vel.y <= 0 and frame >= 9 then
            id = "JUMP_FALL"
        end

        if id == 203 and m.prevAction == ACT_AIR_HIT_WALL then
            id = "FIRSTIE"
        end

        if prevId ~= id and m.playerIndex == 0 then
            prevId = id

            local startFrame
            if id == 4 then
                startFrame = 2
            elseif id == "JUMP_FALL" then
                startFrame = 9
            else
                startFrame = 0
            end

            frame = startFrame
            set_anim_to_frame(m, startFrame)
        end

        smlua_anim_util_set_animation(m.marioObj, "MARIO_ANIM_" .. id)

        if id == "FIRSTIE" and is_anim_past_frame(m, 5) ~= 0 then
            play_sound(SOUND_ACTION_SPIN, m.marioObj.header.gfx.cameraToObject)
        end
    end

    update_mario_body(m)
end

hook_event(HOOK_MARIO_UPDATE, mario_update)

hook_event(HOOK_ON_SET_MARIO_ACTION, function(m)
    if not active_player(m) then
        return
    end

    local mBody = m.marioBodyState
    local d = gMarioDetails[m.playerIndex]

    mBody.torsoAngle.x = 0
    mBody.torsoAngle.y = 0
    mBody.torsoAngle.z = 0

    mBody.headAngle.x = 0
    mBody.headAngle.y = 0
    mBody.headAngle.z = 0

    d.torsoAngle.x = 0
    d.torsoAngle.y = 0
    d.torsoAngle.z = 0
end)